/************************************************************************
 *			finger daemon user-configurable			*
 *			  (c)1999-2001 Robert Trebula R0B0			*
 *		       http://www.elf.stuba.sk/~trebular		*
 ************************************************************************/
/* vim: set cindent: 
   vim: set ts=4: */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <utmp.h>
#include <malloc.h>
#include <time.h>
#include <paths.h>
#include <signal.h>
#include <setjmp.h>
#include "fingerd.h"
#include "config.h"

struct hostent *remote;
jmp_buf jmpbuff;

#ifndef NO_IDENT
char *remuser;
static char *remote_user(struct sockaddr_in *his); //gets name of remote user
#endif /*NO_IDENT*/

int main(int argc, char *argv[]) {
	char name[NAME_LEN];
	fgets(name,NAME_LEN,stdin);
	terminate(name);	//cleaning from the network rubbish
#ifdef MSGFILE
	show_file(MSGFILE);
#endif	/*MSGFILE*/
	check_remote(0);
	if (*name=='\0') {
		show_stats();
	} else if (!strcmp(name,".version")) {
		show_version();
		exit(0);
	}
	else show_user(name);
	return(0);
}

void show_user(char *name) { //for `finger someone@somewhere`
	DATA fs;
#ifndef NO_RC	
	FILE *rc;
#endif /*NO_RC*/
	struct passwd *pwent;
	pwent=getpwnam(name);
	if (!pwent) {
		printf("fingerd: %s -no such user.\n",name);
		return;
	} 
	fs=(struct fstruct *)malloc(sizeof(struct fstruct));
	if (!fs) {
		printf("fingerd:not enough memory\n");
		exit(-1);
	}
	initialize(fs);
	read_pwent(fs,pwent); //reads login, realname,... from /etc/passwd
	stat_mailfolder(fs); //reads mail folder time stamp (last mail reading)
	read_utmp(fs,name); //reads stuff from /var/log/utmp
#ifndef NO_RC
	rc=getrc(pwent->pw_dir);
	if (rc!=NULL) {
		read_rcfile(fs,rc); //reads user's .fingerrc
		fclose(rc);
	}
#endif /*NO_RC*/
	if(fs->isuser&&(!fs->loggedon)) parse_wtmp(fs,name);
	showup(fs);
	free(fs);
}

void showup(DATA fs) { 
	struct ustruct *where;
	char t[101];
	if (!fs->isuser) {
		printf("fingerd: %s -no such user.\n",fs->login);
		return;
	}
	printf("Login:%s \tReal name:%s\n",fs->login,fs->realname);
	printf("Home directory:%s \tShell:%s\n",fs->homedir,fs->shell);
	printf("Mail last read: %s",ctime(&(fs->mailread)));
	if (fs->loggedon) {
		where=fs->utmplist; 
		while (where) {
			strftime(t,100,"%c",localtime(&(where->since)));
			printf("Logged from %s since %s, %s idle.\n",
				where->hostname,t,count_time(where->idle));
			where=where->next;
		}
		printf("Totally %d connection%s.\n",fs->loggedon,fs->loggedon>1?"s":"");
	} else {
		printf("Last logged from %s on %s",fs->lhost,ctime(&(fs->llogged))); //could have used strftime instead
	}
	if (!show_file(fs->plan)) printf("No plan.\n");
	if (!show_file(fs->pgpkey)) printf("No pgp key.\n");
}

void read_utmp(DATA fs,char *name) {
	int fd;
	struct utmp ue;
	struct ustruct *w;
	fd=open(_PATH_UTMP,O_RDONLY);
	if (fd==-1) {
		perror("finger");
		exit(-1);
	}
	while (read(fd,&ue,sizeof(ue))) {
		if (strcmp(ue.ut_user,name)) continue;
		if (get_idle(ue.ut_line)==-1) continue;
		if (fs->utmplist==NULL) {
			fs->utmplist=(struct ustruct *)malloc(sizeof(struct ustruct));
			if(!(fs->utmplist)) {
				printf("fingerd:not enough memory\n");
				exit(-1);
			}
			fs->last=fs->utmplist;
		} else {
			fs->last->next=(struct ustruct *)malloc(sizeof(struct ustruct));
			if(!(fs->last->next)) {
				printf("fingerd:not enough memory\n");
				exit(-1);
			}
			fs->last=fs->last->next;
		}
		fs->last->next=NULL;
		w=fs->last;
		strncpy(w->tty,ue.ut_line,UT_LINESIZE);
		w->pid=ue.ut_pid;
		w->since=ue.ut_time;
		w->idle=get_idle(w->tty);
		*(ue.ut_host)=='\0' ?
			strcpy(w->hostname,"console") : //is this really true?
			strncpy(w->hostname,ue.ut_host,UT_HOSTSIZE);
		fs->loggedon++;
	}
	close(fd);
}

void show_stats() { //for `finger @some.where`
	struct passwd *pe;
	struct fstruct fs;
#ifndef NO_RC
	FILE *rc;
#endif /*NO_RC*/
	struct ustruct *where;
	printf("Login\tReal name\tLogin time\n");
	setpwent();
	while((pe=getpwent())!=NULL) {//this is slow but beautifully short
		initialize(&fs);      //and it looks nice for users
		read_pwent(&fs,pe);
		read_utmp(&fs,pe->pw_name);
#ifndef NO_RC
		rc=getrc(pe->pw_dir);
		if (rc) {
			read_rcfile(&fs,rc);
			fclose(rc);
		}
#endif /*NO_RC*/
		if (fs.loggedon&&fs.isuser) {
			where=fs.utmplist;
			while (where) {
				printf("%s\t%s\t%s",fs.login,fs.realname,
					ctime(&(where->since)));
					//TODO:indent nicer
				where=where->next;
			}
			dealloc_utmplist(&fs);//free() allocated memory
		}
	}
}

void read_pwent(DATA fs,struct passwd *pe) {
	strncpy(fs->login,pe->pw_name,UT_NAMESIZE);
	strncpy(fs->realname,pe->pw_gecos,LINE);
	strncpy(fs->homedir,pe->pw_dir,LINE);
	strncpy(fs->shell,pe->pw_shell,LINE);
	fs->isuser=1;
	sprintf(fs->plan,"%s/.plan",pe->pw_dir);
	sprintf(fs->pgpkey,"%s/.pgpkey",pe->pw_dir);
}

void parse_wtmp(DATA fs,char *name) {
	int fd;
	struct utmp ue;
	fd=open(_PATH_WTMP,O_RDONLY);
	if (fd==-1) {
		perror("finger");
		return;
	}
	while(read(fd,&ue,sizeof(ue))) {
		if (strcmp(ue.ut_name,name)) continue;
		if (!fs->llogged)
			fs->llogged=ue.ut_time;
		if (!*(fs->lhost)) {
			*(ue.ut_host)=='\0' ?
				strcpy(fs->lhost,"console") : //ehm... ???
				strncpy(fs->lhost,ue.ut_host,UT_HOSTSIZE);
		}
	}
}

void check_remote(int fd) {
	struct sockaddr sa;
	struct sockaddr_in *sin;
	struct hostent *he;
	static struct hostent stathe;
	char *numaddr,*addr;
#ifdef GET_USER
#endif /*GET_USER*/
	int ret;
	int namelen;
	sin=(struct sockaddr_in *)&sa;
	namelen=sizeof(sa);
	ret=getpeername(1,&sa,&namelen);
	if (ret||(!namelen)) {
		if (errno==ENOTSOCK) printf("fingerd:not a socket.\n");
		else {
			perror("fingerd:getpeername");
			exit(-1);
		}
		remote=NULL;
		return;
	}
	numaddr=inet_ntoa(sin->sin_addr);
	//TODO: test this timeout
	signal(SIGALRM,alrmhndlr);
	alarm(5);
	if(!setjmp(jmpbuff)) {
			he=gethostbyaddr((char *)&(sin->sin_addr),4,AF_INET);
			if (!he) {
					herror("fingerd:gethostbyaddr");
					remote=NULL;
					alarm(0);
					return;
			}
			memcpy(&stathe,he,sizeof(stathe));
			remote=&stathe;
			addr=he->h_name;
#ifndef NO_IDENT
			remuser=remote_user(sin);
			   printf("You are %s@%s(%s)\n",remuser?remuser:"unknown",
			   addr,numaddr);
#else NO_IDENT
				printf("You are from %s(%s)\n",addr,numaddr);
#endif /*NO_IDENT*/
	} else {
			remote=NULL;
	}
	alarm(0);
	return;
}

void alrmhndlr(int junk) {
		longjmp(jmpbuff,1);
}

void initialize(DATA fs) {
	*(fs->login)='\0';
	*(fs->realname)='\0';
	*(fs->homedir)='\0';
	*(fs->shell)='\0';
	fs->utmplist=NULL;
	fs->last=NULL;
	fs->loggedon=0;
	fs->isuser=0;
	fs->llogged=0;
	*(fs->lhost)='\0';
	fs->mailread=0;
	fs->allwaysin=0;
	*(fs->plan)='\0';
	*(fs->pgpkey)='\0';
}

int show_file(char *fn) {
	FILE *fd;
	int i;
	char buffer[81];
	struct stat st;
	if (stat(fn,&st)) return(0);
	if (!(st.st_mode&S_IROTH)) return(0);
	fd=fopen(fn,"r");
	if (fd==NULL) return(0);
	for (i=1;i<=PLAN_LENGTH;i++) {
		if (fgets(buffer,sizeof(buffer),fd)==NULL) {
			fclose(fd);
			return(1);
		}
		printf("%s",buffer);
	}	
	fclose(fd);
	return(1);
}

#ifndef NO_IDENT
static char *remote_user(struct sockaddr_in *his) {
	int sockfd;
	int ret;
	struct sockaddr_in sin;
	struct servent *service;
	char buffer[LINE];
	static char name[LINE];
	sockfd=socket(AF_INET,SOCK_STREAM,0);
	if (sockfd==-1) {
#ifdef DEBUG		
		printf("fingerd:remote_user:socket:%s\n",sys_errlist[errno]);
#endif /*DEBUG*/
		return(NULL);
	}
	sin.sin_family=AF_INET;
	service=getservbyname("auth","tcp");
	if (!service) {
		printf("fingerd:remote_user:getservbyname:%s\n",sys_errlist[errno]);
		return(NULL);
	}
	sin.sin_port=service->s_port;
	memcpy(&(sin.sin_addr),remote->h_addr,remote->h_length);
	//TODO: test this timeout
	signal(SIGALRM,alrmhndlr);
	alarm(5);
	if(!setjmp(jmpbuff)) {

			if (connect(sockfd,(struct sockaddr *)&sin,sizeof(sin))==-1) {
#ifdef DEBUG
					printf("fingerd:remote_user:connect:%s\n",sys_errlist[errno]);
#endif /*DEBUG*/
					alarm(0);
					return(NULL);
			}
			sprintf(buffer,"%d,%d\n",ntohs(his->sin_port),79);
			/*XXX: and what if it is not on 79?*/
			write(sockfd,buffer,strlen(buffer));
			read(sockfd,buffer,sizeof(buffer)); //hope it won't be splitted :(
			terminate(buffer);
#ifdef DEBUG
			printf("debug:line got from auth:%s\n",buffer);
#endif /*DEBUG*/
			ret=sscanf(buffer,"%*d , %*d : %*s : %*s : %100s",name);
			if((ret)&&ret!=EOF) return(name);
			else {
					alarm(0);
					return(NULL);
			}
	} else {
#ifdef DEBUG
			printf("fingerd:remote_user:timed-out\n");
#endif /*DEBUG*/
			return(NULL);
	}
	alarm(0);
}
#endif /*NO_IDENT*/

void dealloc_utmplist(DATA fs) {
	struct ustruct *us,*usn;
	us=fs->utmplist;
	while(us) {
		usn=us->next;
		free(us);
		us=usn;
	}
}
