/************************************************************************
 *			finger daemon user-configurable			*
 *		       (c) 1999-2001 Robert Trebula R0B0		*
 *		       http://www.elf.stuba.sk/~trebular		*
 ************************************************************************/

#include <ctype.h>
#include <fcntl.h>
#include <pwd.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <unistd.h>
#include <utmp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "fingerd.h"
#include "config.h"

//statics
static short int sc_day(short int is,char *a);
static short int sc_remote(short int is,char *a);
static short int sc_user(short int is,char *a);

char *line(FILE *f) {
	static char line_buff[LINE];
	char *ret;
	do {
		ret=fgets(line_buff,LINE,f);
		if (!ret) {
			return (NULL);
		}
		terminate(line_buff);
		while (*ret==' '||*ret=='\t') ret++;
	} while (!strlen(ret));
	return(ret);
}

#ifndef NO_RC
void read_rcfile(DATA fs,FILE *f) {//TODO: could parse better
	char *l;
	int command;
	short int true=1;
	while((l=line(f))!=NULL) {
		if (l[0]=='#') continue;
		if (l[0]=='\0') continue;
		if (l[0]=='%') { //section
			l=line(f);
			true=section_condition(l);
			continue;
		}
		if (l[0]=='&') {
			l=line(f);
			true=true&&section_condition(l);
			continue;
		}
		if (l[0]=='|') {
			l=line(f);
			true=true||section_condition(l);
			continue;
		}
		if (l[0]=='{') continue;
		if (l[0]=='}') { true=1;continue;}
		if ((l[0]=='.')&&true) {
			command=getcommand(l+1);
			switch (command) {
				case -1: 
					 break;
				case 0: //nouser 
					fs->isuser=0;
					break;
				case 1: //notloggedin
					fs->loggedon=0;
					break;
				case 2: //realname
					l=line(f);
					strncpy(fs->realname,l,LINE);
					break;
				case 3: //login
					l=line(f);
					strncpy(fs->login,l,UT_NAMESIZE);
					break;
				case 4: //homedir
					l=line(f);
					strcpy(fs->homedir,l);
					break;
				case 5: //shell
					l=line(f);
					strcpy(fs->shell,l);
					break;
				case 6:
					noidle(fs);
					break;
				case 7: //lastlogged
					//TODO: some randomization
					l=line(f);
					fs->llogged=get_time_from_word(l);
					break;
				case 8: //from
					l=line(f);
					from_host(fs,l);
					break;
				case 9: //lastfrom
					l=line(f);
					strncpy(fs->lhost,l,UT_HOSTSIZE);
					break;
				case 10: //allwaysin
					fs->allwaysin=1;
					break;
				default: 
			}
		} 
	}
}

int getcommand (char *line) {
	static char *coms[]={"nouser","notloggedin","realname","login",
		"homedir","shell","noidle","lastlogged","from","lastfrom",
		"alwaysin",NULL};
	int i=0;
	while (coms[i]!=NULL) {
		if (!strncasecmp(line,coms[i],strlen(coms[i]))) {
			return(i);
		}
		i++;
	}
	return(-1);
}

FILE *getrc(char *path) {
	char filename[255];
	FILE *f;
	sprintf(filename,"%s/.fingerrc",path);
	f=fopen(filename,"r");
	return(f);
}
#endif /*NO_RC*/

char *terminate(char *string) {
	register char *a;
	a=string;
	while(*a!='\n'&&*a!='\0'&&*a!='\r')a++; //ehm, know a nicer way?
	*a='\0';
	return(string);
}

char *count_time(time_t t) {
	static char string[LINE+1];
	if (t<120) {
		sprintf(string,"%ld seconds",t);
		return(string);
	}
	t/=60;
	if (t<120) {
		sprintf(string,"%ld minutes",t);
		return(string);
	}	
	t/=60;
	if (t<48) {
		sprintf(string,"%ld hours",t);
		return(string);
	}
	t/=24;
	sprintf(string,"%ld days",t);
	return(string);
}	

#ifndef NO_RC
void noidle(DATA fs) {
/* This is stupid. Everytime someone is fingered, he has different idle times.
   I think this is quite a bug. */
	struct ustruct *where;
	where=fs->utmplist;
	srand(time(NULL));
	while(where) {
		where->idle=rand()%120;
		where=where->next;
	}
}
#endif /*NO_RC*/

time_t get_idle(char *tty) { 
	struct stat st;
	int ret;
	char filename[LINE+1];
	sprintf(filename,"/dev/%s",tty);
	ret=stat(filename,&st);
	if(ret==-1) return(-1);
	return(time(NULL)-st.st_atime);
}

#ifndef NO_RC
time_t get_time_from_word(char *word) {
	if (!strcasecmp(word,"day")) return(time(NULL)-(24*60*60));
	if (!strcasecmp(word,"hour")) return(time(NULL)-(60*60));
	if (!strcasecmp(word,"minute")) return(time(NULL)-(60));
	return(0);
}

void from_host(DATA fs,char *hostname) {
	struct ustruct *where;
	if (fs->loggedon) {
		where=fs->utmplist;
		while(where) {
			strncpy(where->hostname,hostname,UT_HOSTSIZE);
			where=where->next;
		}
	}
#ifdef FAKE_LOGIN
	else if (fs->allwaysin) {
		fs->loggedon=1;
		fs->utmplist=(struct ustruct *)malloc(sizeof(struct ustruct));
		fs->last=fs->utmplist;
		fs->last->next=NULL;
		where=fs->utmplist;
		noidle(fs);
		strncpy(where->hostname,hostname,UT_HOSTSIZE);
		where->since=get_time_from_word("day");
	}
#endif /*FAKE_LOGIN*/
}
#endif /*NO_RC*/

void stat_mailfolder(DATA fs) {
	char mail[LINE+1]; //mail folder
	struct stat buf;
	int ret;
	sprintf(mail,"%s/%s",_PATH_MAILDIR,fs->login);
	ret=stat(mail,&buf);
	if (ret) {
#ifdef DEBUG		
		perror("fingerd:stat_mailfolder");
#endif DEBUG
		fs->mailread=0;
	}
	fs->mailread=buf.st_atime;
}

#ifndef NO_RC
short int section_condition(char *line) {
	short int is;
	char *commands[]={"day","remote","user",NULL};
	typedef short int (*f)(short int is,char *a);
	f functions[]={sc_day,sc_remote,sc_user,NULL};
	char w1[26],w2[26],w3[26];
	char *a; //command argument
	int ret,i;
	ret=sscanf(line,"%25s %25s %25s",w1,w2,w3);
	if (ret<2||ret==EOF) return(0);
	else if (ret==2) {a=w2;is=1;}
	else if (ret==3) {
		a=w3;
		if (!strcmp(w2,"is")) is=1;
		else if (!strcmp(w2,"not")) is=0;
		else return(0);
	}
	i=0;
	while(commands[i]) {
		if (!strcmp(w1,commands[i])) {
			return((functions)[i](is,a));
		}
		i++;
	}
	return(0);
}

static short int sc_day(short int is,char *a) {
	char *days[]={"sunday","monday","tuesday","wednesday","thursday",
		"friday","saturday"};
	time_t t;
	int i,j;
	j=-1;
	for (i=0;i<=6;i++) {
		if (!strcmp(a,days[i])) { j=i;break;}
	}
	if (j==-1) return(0);
	t=time(NULL);
	i=localtime(&t)->tm_wday;
	return(short int)(is?i==j:i!=j);
}

static short int sc_remote(short int is,char *a) {
	int i;
	char *addr;
	short int hit=0;
	struct in_addr *ia;
	if (!remote) {
		return(short int)(!is);
	}
	if (!strcmp(a,remote->h_name)) hit=1;
	else {
		i=0;
		while((addr=remote->h_aliases[i])!=NULL) {
			if (!strncmp(a,addr,strlen(addr))) {
				hit=1;
				break;
			}
			i++;
		}
		if(!hit) {
			i=0;
			while(remote->h_addr_list[i]!=NULL) {
				ia=(struct in_addr *)remote->h_addr_list[i];
				addr=inet_ntoa(*ia);
				if (!strncmp(a,addr,strlen(addr))) {
					hit=1;
					break;
				}
				i++;
			}
		}
	}
	return(short int)(is?hit:!hit);
}

static short int sc_user(short int is,char *a) {
#ifdef NO_IDENT
	return(0);
#else /*NO_IDENT*/
	short int i;
	if (!remuser) return(!is);
	i=strcmp(a,remuser);
	return(short int)(i?!is:is);
}
#endif /*NO_IDENT*/
#endif /*NO_RC*/

void show_version(void) {
	printf("Finger daemon user configurable version %s\n",VERSION);
	printf("by Robert Trebula R0B0 <http://www.elf.stuba.sk/~trebular>\n");
}
